/************************************************************************/
/*									*/
/*	INITFNS.C	"Initializer functions"				*/
/*									*/
/*   This file contains many of the functions used to initialize the	*/
/* CRYPTOCard using the Initializer.  It does not contain any of the 	*/
/* actual I/O functions, but it shows the method in which the data	*/
/* should be sent out.  The I/O functions are found in the files 	*/
/* "CARDFNS.C", "FIXFNS.C", and "PORT.C".				*/
/*									*/
/*   (C) - 1991-1996 CRYPTOCard Corporation.  All rights reserved.	*/
/*									*/
/************************************************************************/

static char sccsid[] =
"@(#)initfns.c  v1.0 Copyright 1996 CRYPTOCard Corporation";

/* only deal with version 3.2 and later tokens */
#define VER32 1

/************************************************************************/
/* Header files:							*/
/************************************************************************/
#include <stdio.h>
#include <string.h>
#include "intlzr.h"
#if defined VER32
#include "langv2.h"
#else
#include "langv1.h"
#endif


/************************************************************************/
/* External Variables:							*/
/************************************************************************/
extern int _PortFD;
extern int _Lang;			/* Language used in CRYPTOCard	*/
extern int _ErrorCode;		/* Error return code from fns	*/




/************************************************************************/
int InitializeCard (CRYPTOCard * Card)
/************************************************************************/
/* This function calls another subset of functions that execute the	*/
/* necessary procedures needed to initialize a CRYPTOCard.  The data	*/
/* placed on the CRYPTOCard is taken from the structure pointed to in	*/
/* the parameters.  This structure contains all of the numeric values	*/
/* that a user entering data onto the CRYPTOCard's keypad would punch	*/
/* in.  For a more detailed description of the structure, see the files	*/
/* "INTLZR.C" and "INTLZR.H".  For a sample of how to enter data into	*/
/* the structure, see "MISC.C".						*/
/* This function returns FALSE (0) if, for any reason, the initializa-	*/
/* tion failed.  If it was successful, it returns TRUE (1).		*/
/************************************************************************/
{
  if (!LinkWithCard ())			/* If we can't link up then	*/
    return (_ErrorCode);		/* return code			*/

  printf ("Starting CRYPTOCard initialization.\n"); 

  SendByteToFixture (SOL|LED_INIT);	/* Start the INIT LED a-glowin'	*/
  printf ("Sending Options...\n");
  if (!SendOptions (Card)) {		/* Send the Option characters	*/
    SendByteToFixture (LED_ERROR);
    return (_ErrorCode);		/* If something went wrong...	*/
  }

  printf ("Sending Keys...\n"); 
  if (!SendKeys (Card))	{		/* Send the Keys, check for	*/
    SendByteToFixture (LED_ERROR);
    return (_ErrorCode);		/* If something went wrong...	*/
  }

  printf ("Sending UserID...\n");
  if (!SendUserID (Card)) {		/* This is becoming repetitive.	*/
    SendByteToFixture (LED_ERROR);
    return (_ErrorCode);		/* If something went wrong...	*/
  }

  printf ("Sending PIN...\n"); 
  if (!SendPIN (Card)) {
    SendByteToFixture (LED_ERROR);
    return (_ErrorCode);		/* If something went wrong...	*/
  }

  SendByteToFixture (LED_DONE);		/* Done LED pleases us muchly	*/

  return ((_ErrorCode=0));		/* If it got this far, then the	*/
}					/* card should be programmed.	*/
/************************************************************************/


/************************************************************************/
int LinkWithCard (void)
/************************************************************************/
/* This function attempts to link with the CRYPTOCard currently 	*/
/* inserted in the Initializer.  If it there is no card present, it 	*/
/* will wait until such a contigency has been met.  If the link was 	*/
/* successful, this function returns TRUE (1), otherwise it returns	*/
/* FALSE (0).								*/
/************************************************************************/
{
  int Counter;				/* Looping variable		*/
  int Linked = FALSE;			/* Default return value		*/
  char * BackDoor = "225371";		/* Back door reset sequence	*/
  char Buf[17];				/* Buffer for comparisons	*/

  if (!IsCardIn ()) {			/* If there's no Card in Fixtr.	*/
/*    printf ("Please insert CRYPTOCard into the Initializer.\n"); */
    while (!IsCardIn ());		/* Wait for ever		*/
  }					/* Card is now inserted		*/
  sleep (1);				/* Wait just a second		*/
  SendByteToFixture (SOL);		/* Engage Solenoid and Flash	*/
/*  printf ("Awaiting message from CRYPTOCard.\n"); */
  /* We don't know if the CRYPTOCard is on or off, so try both ways	*/
  for (Counter = 0; Counter < 2; Counter++) {
    SendByteToFixture (SOL|LED_RESET);	/* Light up the RESET LED	*/
    sleep (1);				/* Pause for but a moment	*/
    SendByteToFixture (SOL);		/* Shut RESET LED off		*/
    if (!GotThisBuf ("        ")) 	/* If we didn't get blanks...	*/
      continue;				/* it's not on (may try again)	*/
    SendDec (BackDoor, 6);		/* Send the backdoor sequence	*/
    strcpy (Buf, "        ");		/* 8 spaces			*/
    strcat (Buf, LanguageTable[0][LOCKED]);	/* Concat. "Locked"	*/
    if (GotThisBuf (Buf)) {		/* If it's on the screen...	*/
      Linked = TRUE;			/* Link was successful		*/
      break;				/* Break from loop		*/
    }					/* End of "if" block		*/
  }					/* End of "loop" block		*/
  if (!Linked) {				/* If not linked...		*/
/*    printf ("CRYPTOCard not responding.  Please remove.\n"); */
    _ErrorCode = CRYPT_NO_RSPNS;	/* Set the global error code	*/
    while (!(SendByteToFixture (LED_ERROR)) & 0x01);	/* Fire up the	*/
  }
/*  else
    printf ("CRYPTOCard communication established.\n"); */
  return (Linked);
}
/************************************************************************/




/************************************************************************/
int SendOptions (CRYPTOCard * Card)
/************************************************************************/
/* This function looks for the "Options?" message on the CRYPTOCard's	*/
/* display, and then sends the 3 options defined in the CRYPTOCard	*/
/* structure (pointed to in the parameter) across to the CRYPTOCard.	*/
/* If the "Options?" message is not found, it returns FALSE, otherwise	*/
/* it returns TRUE.							*/
/************************************************************************/
{
  int Counter;					/* That looper again	*/

  if (!SendByte (ENT_KEY)) {	/* Send ENT: gets out of Locked	*/
/*    printf ("CRYPTOCard not responding.  Please remove.\n"); */
    _ErrorCode = CRYPT_NO_RSPNS;	/* Set the error indicator	*/
    while (SendByteToFixture (LED_ERROR) & 0x01)
      return (FALSE);			/* Flash an error and return	*/
  }					/* End the check		*/

  if (!GotThisBuf (LanguageTable[0][OPTIONS])) {/* Look for "Options?"	*/
    _ErrorCode = CRYPT_NO_OPTS;			/* Set error		*/
    return (FALSE);				/* message. If no go...	*/
  }
  for (Counter = 0; Counter < 3; Counter++)	/* Send the 3 options	*/
    if (!(SendOctal (Card->Options[Counter]))) {/* Just do it.		*/
      _ErrorCode = CRYPT_COM_ERR;		/* Uh oh...		*/
      return (FALSE);				/* Communication error.	*/
    }
  return (TRUE);				/* All's well...	*/
}
/************************************************************************/


void GetEncryptedResult (oct *Key, char *Result)
/*
 * This function will encrypt nulls with the Key
 * parameter using the DES algorithm, and place the result	
 * (which is in hexadecimal format) into the buffer pointed to in the 
 * parameters.  The buffer contains the ASCII version of each hex	
 * digit.  The DES algorithm used is supplied by the UNIX crypt functions
 */
{
  char BinKey[64];			/* The ASCII version of the key	*/
  char Zeros[64];			/* The encryption uses zeros	*/
  char Encrypt[17];			/* Copy of the final result	*/
  char Temp[3];				/* Temporary holder of hex val	*/
  int i,j,k;				/* Iteration variable		*/
  unsigned char val, mask;
  unsigned char enc_result[4];
  unsigned char half;

  memset( Result, '\0', 8 );

  /* convert the key to a string of characters (0 or 1) */
  for( i = 0, k = 0 ; i < 8 ; i++ ) 
     { /* for each octal value */
     val = OctDigToASCII( Key[i] ); 
     mask = 0x80;
     for( j = 0 ; j < 8 ; j++ ) 
	{ /* for each bit in a byte */
	BinKey[k++] = ((val & mask) ? 0x01 : 0x00);
	mask = mask >> 1;
	}
     }

  /* set the key */
  des_set_key( BinKey );

  /* encrypt binary zeros */
  memset (Zeros, '\0', 64);		/* Fill with zeros		*/
  des_encrypt( Zeros, 0 );

  /* Zeros contains the encrypted result */
  /* convert the "upper" half of result to a 4 byte binary value */
  for( i = 0, mask = 0x80 ; i < 32 ; i++ )
     { /* for each character */
     if( (i % 8) == 0 )
	{
	enc_result[i/8] = 0;
	mask = 0x80; 
	}
     if( Zeros[i] )
        enc_result[i/8] |= mask;
     mask = mask >> 1;
     }

  /* finally, convert this binary result to ASCII for the caller */
  for( i = 0, j = 0 ; i < 4 ; i++ )
     { /* converts to 8 ASCII characters */
     half = (enc_result[i] & 0xf0) >> 4;
     Result[j++] = (half > 0x09) ? (half-0x09+0x40) : (half+0x30);
     half = (enc_result[i] & 0x0f);
     Result[j++] = (half > 0x09) ? (half-0x09+0x40) : (half+0x30);
     }
  /* null terminate the result */
  Result[j] = '\0';
}



/************************************************************************/
int SendKeys (CRYPTOCard * Card)
/************************************************************************/
/* This function sends the keys defined in the CRYPTOCard structure	*/
/* pointed to in the parameters across to the CRYPTOCard inserted in	*/
/* the Initializer.  It first calculates the number of keys by examin-	*/
/* ing the 3rd character in the 3rd Option (which just happens to be	*/
/* the number of keys), then sends them all across.  If there was a 	*/
/* communication error or the "Keyx?" messages couldn't be found, then	*/
/* this function returns FALSE (0).  Otherwise it returns TRUE (1).	*/
/************************************************************************/
{
  int NumKeys = Card->Options[2][2] - 0x30;	/* Calc. Number of keys	*/
  int Count, Count2;				/* There (s)he is again	*/
  int Decimal = (Card->Options[0][1] == '0') ? FALSE : TRUE;
  int Phone = (Card->Options[0][2] == '0') ? FALSE : TRUE;
  /* The above two lines calculate whether the display is in decimal	*/
  /* format and/or telephone format (see CRYPTOCard DES Challenge Token	*/
  /* Operations and Systems Guide, Revision 2.2, Page 5), as opposed to	*/
  /* the default mode - hexadecimal.					*/
  char Encrypt[20];				/* Holds encrypted rslt	*/

  _Lang = Card->Options[2][1] - 0x30;		/* The language used	*/
  for (Count = 0; Count < NumKeys; Count++) {
    SendByte (ENT_KEY);			/* Move on to next Key	*/
    if (!GotThisBuf (LanguageTable[_Lang][Count])) { /* We got "Keyx?"	*/
      _ErrorCode = CRYPT_NO_KEYX;
      return (FALSE);				/* Nope, Sorry!		*/
    }
    GetEncryptedResult (Card->Keys[Count], Encrypt); /* Get DES result	*/
    if (Decimal)				/* If decimal display..	*/
      for (Count2 = 0; Count2 < 8; Count2++)	/* Go through encrypt.	*/
	if (Encrypt[Count2] >= 'A')		/* If it needs change	*/
	  Encrypt[Count2] -= 0x11;		/* Set it to a digit	*/
    if (Phone)					/* If phone display...	*/
       Encrypt[3] = '-';			/* Make 4th char "-".	*/
    for (Count2 = 0; Count2 < 8; Count2++)	/* Send 8 octal digits	*/
      SendOctal (Card->Keys[Count][Count2]);	/* Send this octal num	*/
    SendByte (ENT_KEY);			/* Send an ENTer next	*/
    /* CRYPTOCard should do an encryption on the key using 00000000,	*/
    /* and send back the result.  We'll compare the CRYPTOCard's to	*/
    /* ours to make sure it is working fine.				*/
    if (!(GotThisBuf (Encrypt))) {		/* Didn't get it back..	*/
/*      printf ("\nEncrypted results do not match.\n"); */
      while (SendByteToFixture (LED_ERROR) & 0x01); /* Lighten up ERROR	*/
      _ErrorCode = CRYPT_BAD_ENCRPT;		/* Set code		*/
      return (FALSE);				/* Indicate error	*/
    }
  }						/* End the NumKeys loop	*/
  return (TRUE);				/* We are happy.	*/
}
/************************************************************************/



/************************************************************************/
int SendUserID (CRYPTOCard * Card)
/************************************************************************/
/* This function first determines whether it needs to send the User ID	*/
/* or not, according to the Options.  If not, it returns a TRUE to 	*/
/* indicate that no errors occurred (even though it didn't read the ID)	*/
/* It then gets an ASCII version of the User's ID to be used when	*/
/* checking that it was entered correctly.  It then sends the octal 	*/
/* version of the ID to the CRYPTOCard and tests the echo to make sure	*/
/* that all is well.  If something has gone amiss, the function returns	*/
/* FALSE, otherwise we get a gleeful TRUE.				*/
/* Note that the User ID should be 8 bytes long.  If it is only 7, a	*/
/* white space character should be appended.				*/
/************************************************************************/
{
  char UserID[9];			/* Buffer to hold the ASCII ID	*/
  int Counter;				/* My friend.			*/

  if (Card->Options[1][0] == '0' || Card->Options[1][0] == '2')	
					/* User ID not required!	*/
    return (TRUE);			/* Get on outa here.		*/

  SendByte (ENT_KEY);		/* Move on to User ID.		*/
  if (!(GotThisBuf (LanguageTable[_Lang][USER_ID]) )) {
    _ErrorCode = CRYPT_NO_ID;		/* Set the code			*/
    return (FALSE);			/* "User ID" not displayed.	*/
  }

  OctToASCII (UserID, Card->UserID, 8);	/* Get an ASCII version.	*/
  for (Counter = 0; Counter < 8; Counter++) {
    SendOctal (Card->UserID[Counter]);	/* Send each octal byte to Card	*/
/*    GetByte (_Port);		*/	/* Get byte num and ignore it	*/
  }
  SendByte (ENT_KEY);		/* Complete the sequence	*/

  if (!(GotThisBuf (UserID) )) {	/* ID doesn't match!		*/
    _ErrorCode = CRYPT_BAD_ID;		/* set value			*/
    return (FALSE);			/* Inform the caller.		*/
  }

  return (TRUE);			/* All is fine and dandy.	*/
}
/************************************************************************/




/************************************************************************/
int SendPIN (CRYPTOCard * Card)
/************************************************************************/
/* This function sends the User's PIN across to the CRYPTOCard.  It	*/
/* then sends it again to satisfy the CRYPTOCard's need for		*/
/* verification.							*/
/************************************************************************/
{
  SendByte (ENT_KEY);		/* Move CRYPTOCard to next step	*/

  if (!(GotThisBuf (LanguageTable[_Lang][NEW_PIN]) )) {
    _ErrorCode = CRYPT_NO_NEWPIN;
    return (FALSE);
  }

  SendDec (Card->UserPIN, 8);		/* Send this PIN across now	*/
  SendByte (ENT_KEY);		/* Finish with ENTer		*/

  if (!(GotThisBuf (LanguageTable[_Lang][VERIFY]) )) {
    _ErrorCode = CRYPT_NO_VER;
    return (FALSE);
  }

  SendDec (Card->UserPIN, 8);		/* Send it yet again		*/
  SendByte (ENT_KEY);		/* Close it off			*/

  if (!(GotThisBuf (LanguageTable[_Lang][CARD_OK]) )) {
    _ErrorCode = CRYPT_NO_OK;
    return (FALSE);			/* Oops				*/
  }

  SendByte (ENT_KEY);		/* Move on to PIN? message	*/

  if (!(GotThisBuf (LanguageTable[_Lang][PIN]) )) {
    _ErrorCode = CRYPT_NO_PIN;
    return (FALSE);
  }

  return (TRUE);			/* Card done successfully	*/
} 					/* So long!			*/
/************************************************************************/
