/*
 *									
 *	INTLZR.C	"Initializer C code"			
 *								
 *   This file contains the mainline function for the Initializer
 * program.  The functions are found in other C modules in this 
 * directory.
 *   INTLZR.H contains the function prototypes as well as various	
 * definitions useful for this application.			
 *
 *   As it stands, this program simply sets up a structure that holds
 * an initialization sequence for a CRYPTOCard.  The infor-	
 * mation is copied into function GetCardInfo().	
 * The structure contains strings of ASCII characters that represent	
 * the actual keystrokes that a user would hit on the CRYPTOCard's 
 * keypad when prompted for the specific information.  Note that these
 * strings are not NULL terminated.  As such, the structure is simply	
 * a continuous block of ASCII characters.  Note also that any special
 * keys that are pressed on the keypad (such as the SCROLL key or the
 * ENT key) are inserted by the functions that actually send the data,
 * and should not be included in the structure.  Examination of the 
 * header file as well as the GetCardInfo() function will make this
 * more clear.							
 *   Once the structure is set up, it is simply a matter of calling the
 * function InitializeCard() to do just that.  All of the functions
 * used by InitializeCard() are included in separate modules in order
 * that you may decipher more concerning the communication protocol
 * with the Initializer.				
 *							
 *   (C) - 1991-1996 CRYPTOCard Corporation.  All rights reserved.
 */


/* The necessary header files are included below:			*/
#include <sys/types.h>
#include <sys/ioctl.h>
#include <unistd.h>
#include <a.out.h>
#include <stdio.h>
#include <fcntl.h>
#include <signal.h>
#include <errno.h>
#include <termios.h>
#include <string.h>
#include <stdlib.h>
#include "intlzr.h"
#include "md5.h"

int port = 0;
int _PortFD = 0;
int _ErrorCode = 0;
int _Lang = 0;
char *progname = "";
int debug_flag = 0;
int rand_flag = 0; /* indicates whether initial challenge, if required, is generated 
		      randomly or obtained from terminal */
struct termios saved_term;
int saved_ldisc = 0;

static char sccsid[] =
"@(#)intlzr.c	1.3 Copyright 1996 CRYPTOCard Corporation";

/* various signal handlers */
void		sig_hup();
static void restore_settings( void );
void            update_one();
void		update_user();
void            update_all();

main (int argc, char **argv)
/************************************************************************/
/* The mainline program follows:					*/
/************************************************************************/
	{
  	int Error;
	int set_rts;
	int i;
	int state;
	FILE *fp;
	CRYPTOCard Card;
	char *init_port, *user;
	char argval;
	progname = *argv++;
	argc--;
	debug_flag = 0;

	signal(SIGHUP, sig_hup);

        user = NULL;
        port = 0;
	while(argc) {

		if(**argv != '-') {
			usage();
		}

		argval = *(*argv + 1);
		argc--;
		argv++;

		switch(argval) {

		case 'p':
			if(argc == 0) {
				usage();
			}
			init_port = *argv;
			port = 1;
			argc--;
			argv++;
			break;

                case 'u':
			if(argc == 0) {
				usage();
			}
                        user = *argv;
			argc--;
			argv++;
			break;

		case 'r':   /* random initial challenge */
			rand_flag = 1;
			break;
		
		case 'x':
			debug_flag = 1;
			break; 

		default:
			usage();
			exit(-1);
		}
	}

        fp = fopen(MASTERKEY_FILE, "r");
        if (fp == NULL) {
            printf("\nError: can't open %s to get common secret\n", MASTERKEY_FILE);
	    exit (-1);
	}
	fclose(fp);

	if (!user) {
	    usage();
	    exit(-1);
	}
	fp = fopen(INIT_FILE, "r");
        if (fp == NULL) {
	    printf("Error: can't open %s\n", INIT_FILE);
	    exit(-1);
        }

	if (port) {
            InitPort( init_port, &saved_term, &saved_ldisc, &_PortFD ); 
            if( atexit( restore_settings ) != 0 )
	        printf( "atexit error: may not be able to restore port settings\n");
        }

        if (strcmp(user, "all")) 
	    update_one(&Card, fp, user);
        else
            update_all(&Card, fp);

	fclose(fp);
        }


#if 0
/************************************************************************/
/* encrypts (crypt_flag = 1) or decrypts (crypt_flag = 0) user's key    */  
/* returns 1 when everything is O.K., otherwise 0 is returned           */
/************************************************************************/
int crypt_key(char *username, char *key, int crypt_flag)
{
   FILE *fp;
   char md5input[MAX_LINE_LEN];
   char md5output[MD5_OUTPUT];
   char *tmp;
   char file_line[MAX_LINE_LEN], secret[MAX_LINE_LEN];
   int line_count = 0;
   int i;

   fp = fopen(MASTERKEY_FILE, "r");
   if (fp == NULL) {
       printf("\nError: can't open %s to get common secret\n", MASTERKEY_FILE);
       return 0;
   }

   secret[0] = '\0';
   do {
       tmp = fgets(file_line, MAX_LINE_LEN, fp);
       if (!tmp) {
           break;
       }
       line_count++;
       if (*tmp == '#')
	   continue;
       if (sscanf(file_line, "%s\n", secret) != 1) {
           printf("Warning: incorrect entry format at line %d in file %s\n", line_count, MASTERKEY_FILE);
           secret[0] = '\0';
	   continue;
       }
   } while (!secret[0]);

   if (!secret) {
       printf("Masterkey is not found in file %s\n", MASTERKEY_FILE);
       return 0;
   }	

   bcopy(secret, md5input, OCT_KEY_LEN);
   strncat(md5input, username, strlen(username));
   md5_calc(md5output, md5input, strlen(md5input));

   if (!crypt_flag) {
       tmp = (char *) malloc(sizeof(char) * MD5_OUTPUT);
       bcopy(key, tmp, MD5_OUTPUT);
   }
   for (i = 0; i < MD5_OUTPUT; i++) {
       key[i] ^= md5output[i];
   }

   bcopy(secret, md5input, OCT_KEY_LEN);
   md5input[strlen(secret)+1] = '\0';
   strncat(md5input, crypt_flag ? key : tmp, MD5_OUTPUT);
   md5_calc(md5output, md5input, strlen(md5input));
   for (i = 0; i < 8; i++) {
       key[MD5_OUTPUT + i] ^= md5output[i];
   }
   key[OCT_KEY_LEN] = '\0';
   fclose(fp);
   return 1;
}
#endif

/************************************************************************/
/* updates the record of a user whose name is stored in the string user */  
/************************************************************************/
void update_user(char *user, char *option1, char *option2, char *option3, char *key, char *userid, char *ipin, CRYPTOCard *card)
{
        int i, len, Error; 
        char tmp_key[49];
	int keyvalue;

        bcopy(option1, card->Options[0], 3);
        bcopy(option2, card->Options[1], 3);
	bcopy(option3, card->Options[2], 3);

        bcopy(key, tmp_key, 48);
	for (i = 0; i < OCT_KEY_LEN; i++) {
	  if(sscanf(&tmp_key[2*i], "%2x", &keyvalue)!=1) {
	    fprintf(stderr,"Bad hex value for user %s: %s\n",&tmp_key[2*i]);
	    return;
	  }
	  key[i]=keyvalue&0xff;
	}
        
        if (!crypt_key(user, key, DECRYPT))
            return;

	for (i = 0; i < 8; i++)
	    bcopy(&key[i*3], card->Keys[0][i], 3);

	if (option2[0] == '1' || option2[0] == '3') {
	    len = strlen(userid);
	    for (i = 0; i < 8; i++)
	        ASCIIToOct(i < len ? userid[i] : SPACE, card->UserID[i]);
	}
	else {
	    int j;
	    for (i = 0; i < 8; i++) {
		for (j = 0; j < 3; j++) {
	            card->UserID[i][j] = SPACE;
		}     
	    } 
	}

	if ((i = strlen(ipin)) < (option2[2] - '0')) {
            printf("Error: IPIN entry for user %s is too short. Initialization failed\n", user);
            return;
	}
        bcopy(ipin, card->UserPIN, strlen(ipin));

	do {
	    if (port) 
		printf("Initialize card for user %s. Press <Enter> when ready.\n", user);
	    else 
		printf("Transfer user %s's record from %s to %s. Press <Enter> when ready.\n", user, INIT_FILE, CC_FILE);
	} while (getchar() == EOF);

	if (port) {
            PrintCardInfo(card);
	    if( (Error = InitializeCard( card )) == 0 )
                printf( "CRYPTOCard initialized successfully\n" );
	    else
   	        printf ("Error return code: %d.\n", Error);
        }

	if (!write_to_ccfile(user, key, option1, option2[0] >= '2'? 1 : 0)) {
	    printf("Error: cannot update %s\n", CC_FILE);
	    return;
        }
        	
	if (!fill_date(user)) {
	    printf("Error: cannot update the date column for user %s\n", user);
        }         
}


/************************************************************************/
/* updates the record of one user in INIT_FILE                          */  
/************************************************************************/
void update_one(CRYPTOCard *card, FILE *fp, char *user)
{
        char file_line[MAX_LINE_LEN], someone[MAX_LINE_LEN], option1[MAX_LINE_LEN], option2[MAX_LINE_LEN], option3[MAX_LINE_LEN], key[MAX_LINE_LEN], userid[MAX_LINE_LEN], ipin[MAX_LINE_LEN], serial[MAX_LINE_LEN], date[MAX_LINE_LEN];
        char *tmp;
	int i, len, Error;
	int line_count = 0;

i = 0;
        /* finds user in INIT_FILE */
	do {
	    tmp = fgets(file_line, MAX_LINE_LEN, fp);
	    if (!tmp) {
		break;
	    }
	    line_count++;
	    if (*tmp == '#')
		continue;
            if ((i = sscanf(file_line,"%s %s %s %s %s %s %s %s %s", someone, option1, option2, option3, key, userid, ipin, serial, date)) != 9 ) {
		printf("Error: incorrect entry format at line %d in file %s. Data on this will be ignored.\n", line_count, INIT_FILE);
		someone[0] = '\0';
		continue;
	    }
        } while (strcmp(someone, user)) ;
	
        if (tmp){
	    update_user(user, option1, option2, option3, key, userid, ipin, card);
        }
	else {
	    printf("Cannot find user %s in %s. Initialization failed.\n", user, INIT_FILE);
	}
}

  
/************************************************************************/
/* updates the record of all users in INIT_FILE                         */  
/************************************************************************/
void update_all(CRYPTOCard *card, FILE *fp)
{
        char file_line[MAX_LINE_LEN], user[MAX_LINE_LEN], option1[MAX_LINE_LEN], option2[MAX_LINE_LEN], option3[MAX_LINE_LEN], key[MAX_LINE_LEN], userid[MAX_LINE_LEN], ipin[MAX_LINE_LEN], serial[MAX_LINE_LEN], date[MAX_LINE_LEN];
	char *tmp;
	int Error;
	int line_count = 0;

        while ((tmp = fgets(file_line, MAX_LINE_LEN, fp))) {
	    line_count++;
	    if (*tmp == '#')
		continue;
        
            if ( sscanf(file_line,"%s %s %s %s %s %s %s %s %s", user, option1, option2, option3, key, userid, ipin, serial, date) != 9 ) {
		printf("Error: incorrect entry format at line %d in file %s. Data on this will be ignored.\n", line_count, INIT_FILE);
		continue;
	    }
	    update_user(user, option1, option2, option3, key, userid, ipin, card);
        }
}

void
sig_hup(sig)
int	sig;
{
	return;
}



usage()
{
	fprintf(stderr, "Usage: %s [ -p port ] -u all/username [-r] [ -x ]\n",progname);
	exit(-1);
}

void restore_settings( void )
{
	/* restore port settings at exit */
	/* reset the line discipline */
	if( ioctl( _PortFD, TIOCSETD, &saved_ldisc ) < 0 )
		printf( "Unable to restore line discipline: %s \n", strerror( errno ) );
	
	/* set the port back to its original settings */
	if( tcsetattr( _PortFD, TCSANOW, &saved_term ) < 0 ) {
		printf( "tcsetattr error restoring port settings: %s\n", strerror( errno ) );
		}
	/* close the file device */
	close( _PortFD );
}

void log_err(str)
char *str;
{
  fputs(str,stderr);
}

