/************************************************************************/
/*									*/
/*	INTLZR.H							*/
/*									*/
/*   This file contains global definitions and function prototypes for	*/
/* the Initializer program.						*/
/*									*/
/*   (C) - 1991-1996 CRYPTOCard Corporation.  All rights reserved.	*/
/*									*/
/************************************************************************/

/***********************************************************************/
/* Uncomment the following line when running on a LINUX platform       */
/* Otherwise, comment the following line out                           */
/***********************************************************************/  
/* #define LINUX 1 */

#ifdef LINUX
#define TTYDISC 0
#define CRTS_IFLOW 0x00020000 /* RTS flow control of input */ 
#endif
/************************************************************************/
/* Commonly used header files :						*/
/************************************************************************/


/************************************************************************/
/* Global Definitions:							*/
/************************************************************************/
#define	DEBUG               /* Debug mode - currently not activated	*/
#define NO_FXTR             /* Indicate no initializer - for testing	*/

#include "acconfig.h"

#define MAX_LINE_LEN    254
#define SPACE           32
#define OCT_KEY_LEN     24 

#define ERROR   0           /* The following are general defines used	*/
#define FALSE   0           /* to make the code more readable		*/
#define TRUE    1
#define NOERROR 1
#define DECRYPT 0
#define ENCRYPT 1

#define MD5_OUTPUT 16

#define NO_LNG  8           /* Number of languages supported		*/
#define PAUSE   9           /* Delay time for messages in 1/18 seconds	*/
#define TM_OUT_MAX  3       /* Maximum time out in seconds		*/

#define LED_NONE	0x00	/* No LEDs on (Initializer front panel)	*/
#define LED_RESET	0x01	/* Reset LED on				*/
#define LED_INSERT	0x02	/* Insert LED on			*/
#define LED_INIT	0x04	/* Init LED on				*/
#define LED_DONE	0x08	/* Done LED on				*/
#define LED_ERROR	0x10	/* Error LED on				*/
#define SOL		0x60	/* Solenoid & flash on			*/

#define SCROLL_KEY	0x0B	/* The code for the SCROLL key		*/
#define ENT_KEY		0x0C	/* The code for ENT key on CRYPTOCard	*/

/* The following defines show which position (in each respective	*/
/* language) the messages are in the LanguageTable defined in the file	*/
/* "LANGUAGE.H".							*/
#define KEY1		0x00
#define KEY2		0x01
#define KEY3		0x02
#define USER_ID		0x03
#define NEW_PIN		0x04
#define VERIFY		0x05
#define CARD_OK		0x06
#define PIN		0x07
#define LOCKED		0x08
#define OPTIONS		0x09

/* The following defines identify the results from the call to the 	*/
/* function InitializeCard(). 						*/
#define CRYPT_NO_RSPNS		-1	/* No response from the Card	*/
#define CRYPT_COM_ERR		-2	/* Communication error		*/
#define CRYPT_NO_OPTS		-3	/* "Options?" message not found	*/
#define CRYPT_NO_KEYX		-4	/* "Keyx?" message not found	*/
#define CRYPT_BAD_ENCRPT	-5	/* Encryption from card doesn't
					   match the computers version	*/
#define CRYPT_NO_ID		-6	/* "User ID?" message not found	*/
#define CRYPT_BAD_ID		-7	/* Displayed ID doesn't match
					   the computer's version	*/
#define CRYPT_NO_NEWPIN		-8	/* "New PIN?" message not found	*/
#define CRYPT_NO_VER		-9	/* "Verify" message not found	*/
#define CRYPT_NO_OK		-10	/* "Card OK" message not found	*/
#define CRYPT_NO_PIN		-11	/* "PIN?" message not found	*/


/************************************************************************/
/* Variable typedefs and structure definitions:				*/
/************************************************************************/
typedef unsigned char byte;	/* Definition of the byte		*/
typedef char oct[3];		/* String of chars representing an	*/
				/* octal byte (ie "163")		*/
typedef char hex[2];		/* String of chars representing a hex	*/
				/* byte (ie "4E") (No null termiator)	*/
typedef char dec;		/* String to represent a single decimal	*/
				/* digit (ie "1" or "8")		*/
typedef struct {		/* Structure contains CRYPTOCard info	*/
   oct	Options[3];		/* Octal digits for all the "Options?"	*/
   oct	Keys[3][8];		/* Octal digits for all of the "Keys"	*/
   oct	UserID[8];		/* Octal digits of the User ID		*/
   dec	UserPIN[8];		/* Decimal format of User's PIN (0-9)	*/
} CRYPTOCard;			/* Aptly named structure		*/

/* NOTES:  The above structure may seem to be oddly designed.  This	*/
/*  implementation is not the most efficient, but it allows for easy	*/
/*  definitions of cards, as one merely has to enter a string that	*/
/*  looks identical to one typed onto the CRYPTOCard's keypad into the	*/
/*  various fields of the structure.  For instance, setting up   	*/
/*  Key1 would be as simple as assigning the string "174" to 		*/
/*	MyCard.Key[0][0];	 			        	*/
/*  where MyCard is declared to be of type CRYPTOCard.  "174" are the	*/
/*  digits that the user would type on the CRYPTOCard's keypad when it	*/
/*  requested "Key1?" and then digit "1".  This ASCII string is the     */
/*  format that the Initializer expects the data (ie sending byte 0x00	*/
/*  is the same as digit 0 on the keypad).				*/

