/************************************************************************/
/*									*/
/*	MISC.C		"Miscellaneous C Functions"			*/
/*									*/
/*   This file contains miscellaneous functions used by the Initializer	*/
/* demonstration program.						*/
/*									*/
/*   (C) - 1991-1996 CRYPTOCard Corporation.  All rights reserved.	*/
/*									*/
/************************************************************************/
 
static char sccsid[] =
"@(#)misc.c  v1.3 Copyright 1996 CRYPTOCard Corporation";
extern int rand_flag; 

/************************************************************************
/* Necessary header files:						*/
/************************************************************************/
#include <stdio.h>
#include <time.h>
#include <signal.h>
#include "intlzr.h"
#include "md5.h"
#include "cryptocard.h"


void PrintCardInfo (CRYPTOCard * Card)
/************************************************************************/
/* This function prints out the information in the CRYPTOCard structure	*/
/* passed in the parameters.  It can be used to double check that all	*/
/* of the proper information is stored in the structure.		*/
/************************************************************************/
{
  int Counter, Counter2, NumKeys;		/* Local variables.	*/

  NumKeys = (int) Card->Options[2][2] - 0x30;	/* The NumKeys is set	*/
						/* by the 3rd digit of	*/
  printf ("Options:\n");			/* Option 3 ([2][2]).	*/
  printf (" %.3s %.3s %.3s\n",			/* Note how the strings	*/
		Card->Options[0],		/* are accessed. Each	*/
		Card->Options[1],		/* Option is an "oct"	*/
		Card->Options[2]);		/* which is defined as:	*/
						/* typedef char oct[3]; */
  printf ("Keys(%d):\n", NumKeys);
  for (Counter2 = 0; Counter2 < NumKeys; Counter2++) {
    for (Counter = 0; Counter < 8; Counter++)
      printf (" %.3s", Card->Keys[Counter2][Counter] );
    printf ("\n");
  }						/* The Keys are similar	*/

  printf ("UserID:\n");
  for (Counter = 0; Counter < 8; Counter++)
    printf (" %.3s", Card->UserID[Counter]);	/* As is the UserID	*/
  printf ("\n");

  printf ("PIN:\n");
  printf (" %.8s\n", Card->UserPIN);		/* The PIN however, is	*/
						/* a string of 8 digits	*/
}
/************************************************************************/


/************************************************************************/
char OctDigToASCII (oct OctalDig)
/************************************************************************/
/* This function converts a single octal digit (represented by a string	*/
/* of 3 ASCII chars - ie "107") into it's ASCII equivalent, and returns	*/
/* this value.  Note that if the Octal number is less than 3 digits	*/
/* long, 0's should be prepended to increase the length to 3.		*/
/************************************************************************/
{
  char ASCIIVal = 0;			/* The final ASCII char		*/

  ASCIIVal = ((OctalDig[0]-0x30)<<6);	/* Add first digit to ASCII	*/
  ASCIIVal |= ((OctalDig[1]-0x30)<<3);	/* Incorporate second digit	*/
  ASCIIVal |= ((OctalDig[2]-0x30));	/* And finally the third	*/

  return (ASCIIVal);			/* and send it back.  See ya.	*/
}


/************************************************************************/
void OctToASCII (char * Buffer, oct * Octal, int Length)
/************************************************************************/
/* This function converts a series of ASCII represented octal digits	*/
/* into an ASCII string.  It stores the string into Buffer.  It starts	*/
/* converting octal digits from the data area pointed to by Octal, and	*/
/* converts Length characters.						*/
/************************************************************************/
{
  for ( ; Length > 0; Length--)			/* Do Length octals 	*/
    *(Buffer++) = OctDigToASCII (*(Octal++));	/* Convert the byte	*/
  *(Buffer) = NULL;				/* NULL terminated	*/
}
/************************************************************************/


/************************************************************************/
/* function that converts an ASCII code into 3 octal digits             */
/************************************************************************/
void ASCIIToOct (char asc_no, char *oct_no)

{
  int i;
  for (i = 0; i < 3; i++) {
      oct_no[2-i] = (asc_no & 0x7) + '0';
      asc_no >>= 3;
  }
}


/************************************************************************/
/* function that copies len characters from string s1 to string s2      */
/************************************************************************/
void bcopy(char *s1, char *s2, int len)
{
  int n;
  if ( (n = len) <= 0) 
      return;
  do
      *s2++ = *s1++;
  while (--n);
}


/************************************************************************/
/* function that gets a line from the terminal and returns its length   */
/************************************************************************/
int get_line(char *line, int max)
{
   if (fgets(line, max, stdin) == NULL)
       return 0;
   else
       return strlen(line); /* note that \n is also counted as a character in fgets() */
}


/************************************************************************/
/* function that gets the initial challenge from the terminal           */
/************************************************************************/
char *get_init(char *init_chal)
{
   int len, i;
   char line[MAX_LINE_LEN];
   len = 0;
   while (len != 9) { /* make sure that input has 9 characters, including the return character \n */
       printf("\nEnter initial challenge (8 hexadecimal digits in the range 00000000 to 99999999)\n");
       if ((len = get_line(line, MAX_LINE_LEN)) != 9) {
           printf("Invalid input\n");	
           continue;
       }
       if (sscanf(line, "%s", init_chal) == EOF) {
           len = 0;
	   printf("Invalid input\n");
	   continue;
       }	
       for (i = 0; i < 8; i++) 
	   if (init_chal[i] < '0' || init_chal[i] > '9') {
	       len = 0;
	       printf("Invalid input\n");
	       break;
	   }
  }
  printf("\n");
  return init_chal;
}


/************************************************************************/
/* Some of the characters strored in the string key may not be          */
/* printable. Instead of printing them directly, the hexadecimal        */
/* equivalent of the ASCII value of each character is printed in the    */
/* file CC_FILE                                                         */ 
/************************************************************************/
int print_enc_key(char *key, FILE *fp)
{
  int i;
  for (i = 0; i < OCT_KEY_LEN; i++) {
      if ((fprintf(fp, "%02x", (unsigned char) key[i])) < 0) {
          printf("\nError: cannot write data in %s\n", CC_FILE);
          return 0;
      }
  }
  return 1;
}


/************************************************************************/
/* updates file CC_FILE, returns 1 when O.K. , 0 otherwise.             */
/************************************************************************/
int write_to_ccfile(char *user, char *key, char *option1, int syn_mode)
{
   FILE *fp;
   off_t user_offset;
   char file_line[MAX_LINE_LEN], someone[MAX_LINE_LEN], somekey[MAX_LINE_LEN], somedisplay[MAX_LINE_LEN], nxtChallenge[MAX_LINE_LEN];
   char *tmp;
   int display;
   int line_count = 0;

   fp = fopen(CC_FILE, "r+");
   if (fp == NULL) {
       if ((fp = fopen(CC_FILE, "w")) == NULL) {
           printf("Error: cannot open or create %s\n", CC_FILE);
           return 0;
       }
       printf("File %s does not exist\n%s is created\n", CC_FILE, CC_FILE);
   }

   /* see if user is already stored in CC_FILE */
   do {
       user_offset = ftell(fp);
       tmp = fgets(file_line, MAX_LINE_LEN, fp);
       if (!tmp) {
	   break;
       }
       line_count++;
       if (*tmp == '#') 
           continue;
       if (sscanf(file_line,"%s %s %s %s", someone, somekey, somedisplay, nxtChallenge) < 3 ) {
	   printf("Warning: incorrect entry format at line %d in file %s\n", line_count, CC_FILE);
	   someone[0] = '\0';
           continue;
       }
   } while (strcmp(user, someone)); 

   display = (option1[1] - '0') * 2 + option1[2] - '0';

   if (!crypt_key(user, key, ENCRYPT))
       return 0;
   if (!tmp) { /* user has not been stored in CC_FILE previously, add it now */  
       if (syn_mode) /* see if synchronous mode is used */
           /* generates random challenge or gets default challenge from terminal */ 
           rand_flag ? random_challenge(nxtChallenge) : get_init(nxtChallenge);
       else
	   memset(nxtChallenge, 32, 8); /* fills string with 8 white spaces */
       nxtChallenge[8] = '\0';
       if ( (fprintf(fp, "%-8s ", user ) < 0) || !print_enc_key(key, fp) || (fprintf(fp, " %4d    %s\n", display, nxtChallenge) < 0) ) {
           printf("Error: cannot write data in %s\n", CC_FILE); 
           return 0;
       }
       fclose(fp);
       return 1;
   }

   /* user has a previous record in CC_FILE */
   if ( fseek(fp, user_offset, SEEK_SET) ) { 
       printf("Can't reseek %s", CC_FILE );
       fclose(fp);
       return 0;
   }
   if (syn_mode) 
       rand_flag ? random_challenge(nxtChallenge) : get_init(nxtChallenge);
   else
       memset(nxtChallenge, 32, 8); /* fills string with 8 white spaces */
   nxtChallenge[8] = '\0';
   if ( (fprintf(fp, "%-8s ", user ) < 0) || !print_enc_key(key, fp) || (fprintf(fp, " %4d    %s\n", display, nxtChallenge) < 0) ) {
       return 0;
   }
   fclose(fp);
   return 1;
}


/************************************************************************/
/* puts date programmed into INIT_FILE                                  */ 
/* returns 1 when O.K. , 0 otherwise                                    */
/************************************************************************/
int fill_date(char *user)
{
   FILE *fp;
   off_t user_offset;
   char file_line[MAX_LINE_LEN], someone[MAX_LINE_LEN], option1[MAX_LINE_LEN], option2[MAX_LINE_LEN], option3[MAX_LINE_LEN], key[MAX_LINE_LEN], userid[MAX_LINE_LEN], ipin[MAX_LINE_LEN], serial[MAX_LINE_LEN], date[MAX_LINE_LEN];
   char *prog_time; 
   char *s, *tmp;
   int line_count;
   int i, len;
   struct tm *localtm;
   time_t *now;
    
   fp = fopen(INIT_FILE, "r+");
   if (fp == NULL) {
       printf("Error: can't open %s\n", INIT_FILE);
       return 0;
   }

   /* finds user whose date field has to be updated */
   do {
       user_offset = ftell(fp);
       tmp = fgets(file_line, MAX_LINE_LEN, fp);
       if (!tmp) {
	   break;
       }
       line_count++;
       if (*tmp == '#') 
           continue;
       if (sscanf(file_line,"%s %s %s %s %s %s %s %s %s", someone, option1, option2, option3, key, userid, ipin, serial, date) != 9 ) {
	   printf("Warning: incorrect entry format at line %d in file %s.\n", line_count, INIT_FILE);
	   someone[0] = '\0';
           continue;
       }
    } while (strcmp(user, someone)); 

   if (tmp) {
       if (!(localtm = (struct tm *) malloc(sizeof(struct tm))) || !(now = (time_t *) malloc(sizeof(time_t))) || !(prog_time = (char *) malloc(sizeof(char) * 13)) ) {
           printf("Error in allocating memory\n");
           return 0;
       }

       /* gets today's date */ 
       time(now);
       localtm = localtime(now);
       strftime(prog_time, 12, "%m/%d/%Y", localtm);
       prog_time[10] = '\0';

       if ( fseek(fp, user_offset, SEEK_SET) ) { 
           printf("Can't reseek %s", INIT_FILE );
           fclose(fp);
           return 0;
       } 
       i = strlen(file_line) - 12;
       for ( s = &file_line[i] ; *s > ' ' ; s-- );

       strcpy( ++s, prog_time );
       if ( fputs(file_line, fp) == EOF ) {
	   printf("Can't rewrite %s\n", INIT_FILE);
	   fclose(fp);
	   return 0;
       }
   }
   else {
       printf("Error: Cannot find user %s in %s.\n", user, INIT_FILE);
       return 0;
   }
   return 1;
}
