/************************************************************************/
/* Note that none of the code available here comes with any kind of     */
/* warranty or support.                                                 */
/*                                                                      */
/*   (C) - 1996 CRYPTOCard Corporation.  All rights reserved.           */
/************************************************************************/


#include "init.h"
#include "md5.h"
#include <sys/time.h>
static char Randpool[16];
static long Randcnt;


/* function to churn the random pool
   obtained from Applied Cryptography, 2nd. Ed. by Bruce Schneier */ 
void churnrand(char *randevent, unsigned int randlen)
{
    MD5_CTX md5;
    MD5Init(&md5);
    MD5Update(&md5, Randpool, sizeof(Randpool));
    MD5Update(&md5, randevent, randlen);
    MD5Final(Randpool, &md5);
}


/* function to generate random bits
   obtained from Applied Cryptography, 2nd. Ed. by Bruce Schneier */
void genrand(char *buf, unsigned int buflen)
{
    MD5_CTX md5;
    char tmp[16];
    unsigned int n;

    while(buflen != 0) {
	MD5Init(&md5);
	MD5Update(&md5, Randpool, sizeof(Randpool));
	MD5Update(&md5, (unsigned char *)&Randcnt, sizeof(Randcnt));
	MD5Final(tmp, &md5);
	Randcnt++;
	n = (buflen < 16) ? buflen : 16;
	memcpy(buf, tmp, n);
	buf += n;
	buflen -= n;
    }
}


/* converts the ASCII equivalent of a character to 3 octal digits */
void ASCIIToOct (unsigned char asc_no, char *oct_no) 

{
  int i;
  for (i = 0; i < 3; i++) {
      oct_no[2-i] = (asc_no & 0x7) + '0';
      asc_no >>= 3;
  }
}


int get_line(char *line, int max)
{
   long tmp;
   char randevent[2];
   int i;
   struct timeval *tp;
   struct timezone *tz;
   tp = (struct timeval *)malloc(sizeof(struct timeval));
   tz = (struct timezone *)malloc(sizeof(struct timezone));

   gettimeofday(tp,tz);
   tmp = tp->tv_usec; /* get the microsecond portion of tp before getting input*/

   if (fgets(line, max, stdin) == NULL)
       return 0;
   else {
       gettimeofday(tp,tz);
       /* compute the difference in the microsecond part before and after getting input*/
       (tp->tv_usec > tmp) ? (tmp = tp->tv_usec - tmp) : (tmp -= tp->tv_usec);
       for (i = 0; i < 2; i++) {
            randevent[1 - i] = (tmp >> i * 8) & 0xFF;
       }
       churnrand(randevent, 2); /* use 16 bits of tmp to churn the random pool */
       return strlen(line); /* note that \n is also counted as a character in
			       fgets() */
   }
}


void get_username(init_record *i_record)

{
   char line[MAX_LINE_LEN];
   int len = 0;

   while (len < 2 || len > 9) { /* make sure that input has 1 to 8 characters,                                     not including the return character \n */   
       printf("\nEnter username (1 to 8 characters)\n");
       len = get_line(line, MAX_LINE_LEN);
       if (sscanf(line, "%s", i_record->username) == EOF) {
	   len = 0;
       }
   }
}


void get_option1(init_record *i_record)

{  int len = 0; 
   char line[MAX_LINE_LEN];
       
       while (len != 4) {
           printf("\nEnter 3 digits for option field 1 [(0-3), (0-1), (0-1)]\n");
           if ((len = get_line(line, MAX_LINE_LEN)) != 4) 
	       continue;
           if (sscanf(line, "%s", i_record->option1) == EOF) {
	       len = 0;
	       continue;
           }
           if (i_record->option1[0] < '0' || i_record->option1[0] > '3' || i_record->option1[1] < '0' || i_record->option1[1] > '1' || i_record->option1[2] < '0' || i_record->option1[2] > '1') {
     	       printf("Invalid Input\n");
	       len = 0;
           }
       } 
}


void get_option2(init_record *i_record)

{  int len = 0; 
   char line[MAX_LINE_LEN];

       while (len != 4) {
           printf("\nEnter 3 digits for option field 2 [(0-3), (0-7), (0, 3-7)]\n");
           if ((len = get_line(line, MAX_LINE_LEN)) != 4) 
	       continue;
           if (sscanf(line, "%s", i_record->option2) == EOF) {
	       len = 0;
	       continue;
           }
           if (i_record->option2[0] < '0' || i_record->option2[0] > '3' || i_record->option2[1] < '0' || i_record->option2[1] > '7' || (i_record->option2[2] < '3' && i_record->option2[2] != '0') || i_record->option2[2] > '7') {
     	       printf("Invalid input\n");
	       len = 0;
           }
       } 
}


void get_option3(init_record *i_record)

{  int len = 0; 
   char line[MAX_LINE_LEN];

       while (len != 4) {
           printf("\nEnter 3 digits for option field 3 [(0-1), (0-7), (1)]\n");
           if ((len = get_line(line, MAX_LINE_LEN)) != 4) 
	       continue;
           if (sscanf(line, "%s", i_record->option3) == EOF) {
	       len = 0;
	       continue;
           }
           if (i_record->option3[0] < '0' || i_record->option3[0] > '1' || i_record->option3[1] < '0' || i_record->option3[1] > '7' || i_record->option3[2] != '1' ) {
     	       printf("Invalid input\n");
	       len = 0;
           }
       } 
}


void get_key(init_record *i_record)

{
   int i, j;
   int len = 0; 
   char line[MAX_LINE_LEN];
   char s[MAX_LINE_LEN];

   do {
       printf("\nGenerate random key for user %s? [y/n]\n", i_record->username);
   } while ((get_line(line, MAX_LINE_LEN) < 2) || (sscanf(line, "%2s", s) == EOF)) ; 
   
   if (!strcmp(s, "y")) {
	char buf[8];
	int k;

        genrand(buf, 8);
	j = 0;
	for (i = 0; i < 8; i++) {
            ASCIIToOct(buf[i],  &(i_record->key[j]));
	    for (k = 0; k < 3; k++) {
		printf("%c", i_record->key[j++]);
	    }
	    printf("\n");
        }
   }
   else {
   	i = 0;
   	j = 0;
   	printf("\nEnter the most significant byte first, then the second most, ...");
   	do {
       	    len = 0;
       	    while (len != 4) { 
	        printf("\nEnter key byte %d in 3 octal digits (000 to 377)\n", j+1);
                if ((len = get_line(line, MAX_LINE_LEN)) != 4) {
                    printf("Invalid input\n");
	            continue;
	        }
                if (sscanf(line, "%s", &(i_record->key[i])) == EOF) {
                    printf("Invalid input\n");
	            len = 0;
	            continue;
                }
                if (i_record->key[i] < '0' || i_record->key[i] > '3' || i_record->key[i+1] < '0' || i_record->key[i+1] > '7' || i_record->key[i+2] < '0' || i_record->key[i+2] > '7') {
                    printf("Invalid input\n");
	            len = 0;
                }
            }
            i+=3;
            j++;
        } while (j < 8); 
   }
}


void get_userid(init_record *i_record)

{
   int len = 0;
   char line[MAX_LINE_LEN];

   if (i_record->option2[0] == '1' || i_record->option2[0] == '3') {
       while (len < 2 || len > 9) {
           printf("\nEnter userid (1 to 8 characters)\n");
           len = get_line(line, MAX_LINE_LEN);
           if (sscanf(line, "%s", i_record->userid) == EOF) {
	       len = 0;
           }
       }
   }
   else {
       printf("\nNo userid required\n");
       strcpy(i_record->userid, "N/A");
   }
}


void get_ipin(init_record *i_record)

{
   int len = 0;
   int min_pin_len = i_record->option2[2] - '0'; 
   char line[MAX_LINE_LEN];

   while (len < min_pin_len+1 || len > 9) {
       printf("\nEnter pin number (%d to 8 characters long)\n", min_pin_len);
       len = get_line(line, MAX_LINE_LEN);
       if (sscanf(line, "%s", i_record->ipin) == EOF) {
	   len = 0;
       }
   }
}


void get_serial(init_record *i_record)

{
   int len = 0;
   char line[MAX_LINE_LEN];

   while (len < 2 || len > 10) {
       printf("\nEnter serial# (1 to 9 characters)\n");
       len = get_line(line, MAX_LINE_LEN);
       if (sscanf(line, "%s", i_record->serial) == EOF) {
	   len = 0;
       }
   }
}


void get_date(init_record *i_record)

{
   strcpy(i_record->date, "       N/A");
}


void bcopy(char *s1, char *s2, int len)
{
  int n;
  if ( (n = len) <= 0)
      return;
  do
      *s2++ = *s1++;
  while (--n);
}


/* Some of the characters stored in the string key may not be printable. */
/* Instead of printing them directly, the hexadecimal equivalent of the  */
/* ASCII value of each character is printed in the file INIT_FILE.       */
int print_enc_key(char *key, FILE *fp)
{
  int i;
  for (i = 0; i < OCT_KEY_LEN; i++) {
      if ((fprintf(fp, "%02x", (unsigned char) key[i])) < 0) {
          printf("\nError: cannot write data in %s\n", INIT_FILE);
          return 0;
      }
  }
  return 1;
}


#if 0
/* encrypts (crypt_flag = 1) or decrypts (crypt_flag = 0) user's key */
/* returns 1 if everything is O.K., otherwise 0 is returned          */
int crypt_key(char *username, char *key, int crypt_flag)

{
   FILE *fp;
   char md5input[MAX_LINE_LEN];
   char md5output[MD5_OUTPUT];
   char *tmp;
   char file_line[MAX_LINE_LEN], secret[MAX_LINE_LEN];
   int line_count = 0;
   int i;

   fp = fopen(MASTERKEY_FILE, "r");
   if (fp == NULL) {
       printf("\nError: can't open %s to get common secret\n", MASTERKEY_FILE);
       return 0;
   }

   secret[0] = '\0';
   do {
       tmp = fgets(file_line, MAX_LINE_LEN, fp);
       if (!tmp) {
	   break;
       }
       line_count++;
       if (*tmp == '#')
	   continue;
       if (sscanf(file_line, "%s\n", secret) != 1) {
	   printf("Warning: incorrect entry format at line %d in file %s\n", line_count, MASTERKEY_FILE);
	   secret[0] = '\0';
	   continue;
       }
   } while (!secret[0]);
  
   if (!secret) {
       printf("Masterkey is not found in file %s\n", MASTERKEY_FILE);
       return 0;
   }

   bcopy(secret, md5input, OCT_KEY_LEN);
   strncat(md5input, username, strlen(username));
   md5_calc(md5output, md5input, strlen(md5input));

   if (!crypt_flag) {
       tmp = (char *) malloc(sizeof(char) * MD5_OUTPUT);	
       bcopy(key, tmp, MD5_OUTPUT);
   }
   for (i = 0; i < MD5_OUTPUT; i++) {
       key[i] ^= md5output[i];
   }

   bcopy(secret, md5input, OCT_KEY_LEN);
   md5input[strlen(secret)+1] = '\0';
   strncat(md5input, crypt_flag ? key : tmp, MD5_OUTPUT);
   md5_calc(md5output, md5input, strlen(md5input));
   for (i = 0; i < 8; i++) {
       key[MD5_OUTPUT + i] ^= md5output[i];
   }
   fclose(fp);
   return 1;
}
#endif


int write_initfile(init_record *i_record)

{
   FILE *fp;
   off_t user_offset;
   char file_line[MAX_LINE_LEN], someone[9], s_option1[4], s_option2[4], s_option3[4], s_key[49], s_userid[9], s_ipin[9], s_serial[10], s_date[9];
   char *tmp;
   int i;
   int line_count = 0;

   fp = fopen(INIT_FILE, "r+");
   if (fp == NULL) {
       if ((fp = fopen(INIT_FILE, "w")) == NULL) {
           printf("\nError: can't open or create %s\n", INIT_FILE);
           return 0;
       }
       printf("\nFile %s does not exist\n%s is created\n", INIT_FILE, INIT_FILE);
   }

   do {
       user_offset = ftell(fp);
       tmp = fgets(file_line, MAX_LINE_LEN, fp);
       if (!tmp) {
	   break;
       }
       line_count++;
       if (*tmp == '#')
	   continue;
       if (sscanf(file_line, "%s %s %s %s %s %s %s %s %s\n", someone, s_option1, s_option2, s_option3, s_key, s_userid, s_ipin, s_serial, s_date) != 9) {
	   printf("Warning: incorrect entry format at line %d in file %s\n", line_count, INIT_FILE);
	   someone[0] = '\0';
	   continue;
       }
   } while (strncmp(i_record->username, someone, strlen(i_record->username)));

    if (!crypt_key(i_record->username, i_record->key, ENCRYPT))
        return 0;

   /* new username */
   if (!tmp) {
       if ( (fprintf(fp, "%-8s %3s %3s %3s ", i_record->username, i_record->option1, i_record->option2, i_record->option3) < 0) || !print_enc_key(i_record->key, fp) || (fprintf(fp, " %-8s %-8s %-9s %10s\n", i_record->userid, i_record->ipin, i_record->serial, i_record->date) < 0) ) {
	   printf("\nError: cannot write data in %s\n", INIT_FILE);
           return 0;
       }
        fclose(fp);
       return 1;
   }

   /* username already exists in file */
   if (fseek(fp, user_offset, SEEK_SET)) {
       printf("\nCan't reseek %s\n", INIT_FILE);
       fclose(fp);
       return 0;
   }
   if ( (fprintf(fp, "%-8s %3s %3s %3s ", i_record->username, i_record->option1, i_record->option2, i_record->option3) < 0) || !print_enc_key(i_record->key, fp) || (fprintf(fp, " %-8s %-8s %-9s %10s\n", i_record->userid, i_record->ipin, i_record->serial, i_record->date) < 0) ) {
	   printf("\nError: cannot write data in %s\n", INIT_FILE);
           return 0;
       }
   fclose(fp);
   return 1;
}
