static const char rcs_id[] = "$Id: cryptocard.c,v 1.1 1995/05/15 17:55:44 karl Exp $";

/*
 *   cryptocard.c
 *	 Uses the challenge and key passed in and determines if the
 *   passed in response is correct or not.  If the response is valid, return
 *   0 else return 1.  If decimal only results are being used, set hexflag
 *   to 0 otherwise pass a 1.  If the response string has a '-' at the
 *   4th character position, the response is interpreted in telephone
 *   display format.
 */
# include <stdio.h>
# include <ctype.h>
# include <string.h>
#include <sys/types.h>
#include <sys/time.h>

#ifdef RADIUS
#include "radius.h"
#endif
# include "cryptocard.h"
# include "des.h"
# define BinToA(c)		((c) <= 9 ? (c) + '0' : ((c) - 10 + 'A'))

int crypt_key();

/* for debugging messages */
extern	int	debug_flag;

# ifdef __STDC__
static void swap(unsigned char *p)
# else /* __STDC__ */
static void swap(p)
unsigned char *p;
# endif /* __STDC__ */
    {
# ifdef i386
# define SWAP_BYTES(a, b)
# else /* i386 */
# define SWAP_BYTES(a, b)	c = a; a = b; b = c
    unsigned char c;
# endif /* i386 */

    SWAP_BYTES(p[0], p[3]); SWAP_BYTES(p[1], p[2]);
    SWAP_BYTES(p[4], p[7]); SWAP_BYTES(p[5], p[6]);
    }

int dump_key(cypher,len,file,line)
unsigned char *cypher;
int len;
char *file;
int line;
{
  short i;
  
  fprintf(stderr,"\n%s:%d ",file,line);
  for(i=0; i<len; i++) {
    fprintf(stderr,"%c%02x ",(isprint(cypher[i]) ? cypher[i] : '.'),cypher[i]&0xff);
  }
  putc('\n',stderr);
}
#define DUMP_KEY(foo,len) dump_key((foo),len,__FILE__,__LINE__)
  

# ifdef __STDC__
char *generate_response(char *key, char *challenge, char *nextchallenge)
# else /* __STDC__ */
char *generate_response(key, challenge, nextchallenge)
char *key;
char *challenge;
char *nextchallenge;
# endif /* __STDC__ */
{
  unsigned char c;
  unsigned long k[2];		/* the key as 8 bytes */
  unsigned char data[8];	/* the data to encrypt */
  unsigned char rb[8];	/* the encrypted result */
  static char resps[9];	/* the calculated response as a string */
  short i;			/* counter */
  des_key_schedule schedule;	/* Processed DES key */

  /* parse the key from the input string */
  {
    unsigned char *kc = (char *)k;
    int temp[8];

    if (key[0] == '0' && key[1] == 'x') {
      /* key is in hexadecimal */

      key+=2;
      sscanf(key, "%2x%2x%2x%2x%2x%2x%2x%2x",
	     &temp[0], &temp[1], &temp[2], &temp[3],
	     &temp[4], &temp[5], &temp[6], &temp[7]);
    }
    else
      {
        /* key is in octal digits (as entered to cryptocard) */

	if(sscanf(key, "%3o%3o%3o%3o%3o%3o%3o%3o",
		  &temp[0], &temp[1], &temp[2], &temp[3],
		  &temp[4], &temp[5], &temp[6], &temp[7])!=8) {
	  log_err("Bad ccsecret input\n");
	  abort();
	}
      }

#ifdef CONSTANT_CHAL
    /* initialize to same */
    kc[0]=0271;
    kc[1]=0024;
    kc[2]=0202;
    kc[3]=0074;
    kc[4]=0374;
    kc[5]=0034;
    kc[6]=0244;
    kc[7]=0322;
#else
    for (i = 0; i < 8; i++)
      kc[i] = temp[i];
#endif
    for(i=0;i<8;i++) {
      printf("Chal %d: %02x %0o\n",i,challenge[i],(unsigned)kc[i]);
    }

  }

  
  /* initialize data to encrypt */
  bcopy(challenge, data, 8);

  /* initialize the key for 16 bit algorithm */
  des_set_key((des_cblock *)k, schedule);

  swap(data);
  memcpy(rb,data,8);
  des_encrypt((unsigned long *)rb, schedule, 1);
  swap(rb);

  for(i=0;i<8;i++) {
    printf("Resp %d: %02x\n",i,rb[i]);
  }

  /* derive the "next challenge" from the current calculated response */
  if (nextchallenge)
    {
      for (i = 0; i < 8; i++)
	{
	  *nextchallenge = rb[i] & 0x0f;

	  if (*nextchallenge >= 10)
	    *nextchallenge -= 10;

	  *nextchallenge++ += '0';
	}

      *nextchallenge = 0;	/* terminating null */
    }
   
  /* convert the first 4 bytes of result to a string */
  bzero(resps, 9);

  for (i = 0 ; i < 4 ; i++)
    {
      resps[i * 2] = BinToA((rb[i] >> 4) & 0xf);
      resps[(i * 2) + 1] = BinToA(rb[i] & 0xf);
    }

  return resps;
}

# ifdef __STDC__
int check_response(char *key, char *challenge, 
		   char *response, int flag,
		   char *nextchallenge)
# else /* __STDC__ */
int check_response(key, challenge, response, flag, nextchallenge)
char *key;
char *challenge;
char *response;
char flag;
char *nextchallenge;
# endif /* __STDC__ */
    {
    char *digits = "0123456789";
    char *res = response;

    while (*res)
	*res++ = toupper(*res);

    res = generate_response(key, challenge, nextchallenge);

    /* make it all decimal if requested */
    if ( flag == '2' || flag == '3')
	{
	char *p = res;

	/* map hex characters 'A' thru 'F' as 0 thru 5 respectively */
	while (*p) {
	    if( *p >= 'A' )
		*p = digits[ *p - 'A' ];
	    else
	    	*p = digits[ *p - '0' ];
	    p++;
	    }
	}

    /* handle telephone format results */
    if (flag == '1' || flag == '3')
	res[3] = '-';
    DEBUG("user's response: %s, expected response: %s\n", response, res); 
    return (!strcmp(response, res));
    }

# ifdef __STDC__
char *random_challenge(char *challenge)
# else /* __STDC__ */
char *random_challenge(challenge)
char *challenge;
# endif /* __STDC__ */
    {
    unsigned long k[2];		/* key */
    unsigned char data[8];	/* the data to encrypt */
    time_t timer;		/* current time counter */
    unsigned char rb[8];	/* result */
    short i;			/* loop counter */
    des_key_schedule schedule;	/* Processed DES key */

    time(&timer);
    data[1] = 0L;
    data[0] = timer;

    /* initialize the key for 16 bit algorithm using whatever is on the stack */
    des_set_key((des_cblock *)k, schedule);

    swap(data);
    memcpy(rb,data,8);
    des_encrypt((unsigned long *)rb, schedule, 1);
    swap(rb);

    /* convert the first 4 bytes of result to a string */
    bzero(challenge, 9);

    for (i = 0 ; i < 4 ; i++)
	{
	challenge[i * 2] = BinToA((rb[i] >> 4) & 0xf);
	challenge[(i * 2) + 1] = BinToA(rb[i] & 0xf);
	}

    for (i = 0 ; i < 9 ; i++)
	/* A through F get converted to 0 through 5 respectively */
	if (challenge[i] > '9')
	    challenge[i] = (challenge[i] - 'A') + '0';

    DEBUG("random challenge = %s\n", challenge);
    return (challenge);
    }

/*
 * cc_getchal opens the cryptocard database file in *filename.  It searches for
 * *username and gets the next challenge from the file (when syn_mode = 1). If
 * syn_mode = 0, a random challenge is generated. Return 0 if successful,
 * posts a message to the system errlog if an error occurs
*/

#ifdef __STDC__
int cc_getchal( char *filename, char *username, char *nxtChallenge, int syn_mode )
# else /* __STDC__ */
int cc_getchal( filename, username, nxtChallenge, syn_mode )
char *filename,
char *username,
char *nxtChallenge,
int syn_mode
# endif /* __STDC__ */
	{
    	char file_line[MAX_LINE_LEN + 1];
    	char tmp[MAX_LINE_LEN], key[MAX_LINE_LEN], response[MAX_LINE_LEN], challenge[MAX_LINE_LEN], display[MAX_LINE_LEN];
    	char buffer[MAX_LINE_LEN];
	off_t user_offset;
   	FILE* fp;
	int result;
	char *s;
	int i;
	unsigned int keybytes;
	char tmp_key[49];

    	DEBUG( "cc_getchal: Looking for %s in %s\n", username, filename );

        tmp[0] = '\0';
	/* Open the database file containing CRYPTOCard keys read write. */
    	fp = fopen(filename, "r+");
    	if ( fp == NULL ) {
		sprintf(buffer,"cc_getchal: Can't open %s\n", filename);
		DEBUG( buffer );
		log_err( buffer );
		return(-1);
    	}
    	/* Now read the file and find the user name */
    	do {
		challenge[0] = '\0';
		user_offset = ftell(fp);
		s = fgets(file_line, MAX_LINE_LEN, fp);
		if ( feof(fp) )
	    		break;
		if ( *s == '#' )
	    		continue;
		result = sscanf(file_line, "%s %s %s %s", tmp, key, display, challenge );
                if (tmp) {
		    bcopy(key, tmp_key, 48);
                    for (i = 0; i < OCT_KEY_LEN; i++) {
		      if(sscanf(&tmp_key[2*i], "%2x", &keybytes)!=1) {
			abort();
		      }
		      key[i]=keybytes&0xff;
		    }
		    
                    if (!crypt_key(tmp, key, DECRYPT))
	                return(-1);
		    DEBUG( "cc_getchal: Found %s with key = %08x%08x%08x%08x%08x%08x and challenge = %s\n", tmp,
			*((int *)&key[0]),
			*((int *)&key[4]),
			*((int *)&key[8]),
			*((int *)&key[12]),
			*((int *)&key[16]),
			*((int *)&key[20]),
		    challenge);	
                }

		if ( result != 4 && result != 3)
	    		tmp[0] = '\0';
    	} while (strcmp(username, tmp));
    
	if ( strcmp(tmp, username) ) {	/* No key for user */
		sprintf(buffer, "cc_getchal: Nothing in %s or missing challenge for %s\n",
	       		filename, username );
		DEBUG( buffer );
		log_err( buffer );
		fclose(fp);
		return(-1);
    	}
	/* copy the next challenge to the user parameter */
	strncpy(nxtChallenge, (syn_mode)? challenge : random_challenge(challenge), 8);
   	nxtChallenge[8] = '\0';
	DEBUG( "cc_getchal: Challenge user %s with %s\n", username, nxtChallenge );
	/* store the next challenge in the cryptocard file */
       	if ( fseek(fp, user_offset, SEEK_SET) ) {
	    log_err( "Can't reseek %s\n", filename );
	    fclose(fp);
	    return(-1);
    	}
    	i = strlen(file_line) - 10;
    	for ( s = &file_line[i] ; *s > ' ' ; s-- ); 

    	strcpy( ++s, nxtChallenge );
    	if ( fputs(file_line, fp) == EOF ) {
	    log_err( "Can't rewrite %s\n", filename );
	    fclose(fp);
	    return(-1);
    	}
	/* close the file and return */
	fclose(fp);
    	return(0);
    }


/* 
 * cc_checkresp compares the response passed in to the response expected 
 *
 */


#ifdef __STDC__
int cc_checkresp( char *filename, char *username, char *response )
# else /* __STDC__ */
int cc_checkresp( filename, username, response )
char *filename,
char *username,
char *response,
# endif /* __STDC__ */
	{
    	char file_line[MAX_LINE_LEN + 1];
    	char tmp[MAX_LINE_LEN], key[MAX_LINE_LEN], challenge[MAX_LINE_LEN], display[MAX_LINE_LEN], *s;
    	char buffer[MAX_LINE_LEN], nxtChallenge[MAX_LINE_LEN];
	off_t user_offset;
   	FILE* fp;
	int result, i, temp;
	int keyvalue;
	char tmp_key[49];

    	DEBUG( "cc_checkresp: Looking for %s in %s\n", username, filename );

        tmp[0] = '\0';    	
	/* Open the database file containing CRYPTOCard keys read write. */
    	fp = fopen(filename, "r+");
    	if ( fp == NULL ) {
		sprintf(buffer,"cc_checkresp: Can't open %s:\n", filename);
		DEBUG( buffer );
		log_err( buffer );
		return(-1);
    	}
    	/* Now read the file and find the user name */
    	do {
		user_offset = ftell(fp);
		s = fgets(file_line, MAX_LINE_LEN, fp);
		if ( feof(fp) )
	    		break;
		if ( *s == '#' )
	    		continue;
		result = sscanf(file_line, "%s %s %s %s", tmp, key, display, challenge );

		if ( result != 4 && result != 3)
	    		tmp[0] = '\0';
    	} while (strcmp(username, tmp));
    
	if ( strcmp(tmp, username) ) {	/* No key for user */
		snprintf(buffer, sizeof(buffer), "cc_checkresp: Nothing in %s or missing key for %s\n",
	       		filename, username );
		DEBUG( buffer );
		log_err( buffer );
		fclose(fp);
		return(-1);
    	}

	bcopy(key, tmp_key, 48);
        for (i = 0; i < OCT_KEY_LEN; i++) {
	  if(sscanf(&tmp_key[2*i], "%2x", &keyvalue)!=1) {
	    snprintf(buffer,sizeof(buffer),"Bad hex value in file %s user %s: %s\n",
		     filename,username,&tmp_key[2*i]);
	    DEBUG(buffer);
	    log_err(*buffer);
	    fclose(fp);
	    return(-1);
	  }
	  key[i]=keyvalue&0xff;
	}
        if (!crypt_key(tmp, key, DECRYPT))
            return(-1);


	/* ensure the database information is terminated properly */
        tmp[25] = '\0';
        display[2] = '\0';
	challenge[8] = '\0';
	response[8] = '\0';
	memset( nxtChallenge, '\0', 9 );
	DEBUG( "cc_checkresp: Checking %s: key=%08x%08x%08x%08x%08x%08x, display=%s, challenge=%s (resp received=%s)\n", username, 
		*((int *)&tmp[0]),
		*((int *)&tmp[4]),
		*((int *)&tmp[8]),
		*((int *)&tmp[12]),
		*((int *)&tmp[16]),
		*((int *)&tmp[20]),
		display, challenge, response );

	temp = check_response( key, challenge, response, display[0], (result == 4) ? nxtChallenge : NULL);

	/* store the next challenge in the cryptocard file */
       	if ( fseek(fp, user_offset, SEEK_SET) ) {
	    log_err( "Can't reseek %s\n", filename );
	    fclose(fp);
	    return(-1);
    	}
    	i = strlen(file_line) - 10;
    	for ( s = &file_line[i] ; *s > ' ' ; s-- ); 
    	strcpy( ++s, nxtChallenge );
    	if ( fputs(file_line, fp) == EOF ) {
	    log_err( "Can't rewrite %s\n", filename );
	    fclose(fp);
	    return(-1);
    	}

	if( !temp ) { 
		sprintf(buffer, "cc_checkresp: Incorrect response for %s (resp recd: %s)\n",
	       		username, response );
		DEBUG( buffer );
		log_err( buffer );
		fclose(fp);
		return(-1);
  	} 
	/* close the file and return */
	fclose(fp);
    	return(0);
    }


/* encrypts (crypt_flag = 1) or decrypts (crypt_flag = 0) user's key */
/* returns 1 if everything is O.K., otherwise 0 is returned          */
#ifdef __STDC__
int crypt_key(char *username, char *key, int crypt_flag)
# else /* __STDC__ */
int crypt_key(username, key, crypt_flag)
char *username,
char *key,
int crypt_flag
# endif /* __STDC__ */
{
   FILE *fp;
   char md5input[MAX_LINE_LEN];
   char md5output[MD5_OUTPUT];
   char *tmp;
   char file_line[MAX_LINE_LEN], secret[MAX_LINE_LEN];
   int line_count = 0;
   int i;
   
   if(debug_flag) {
     if(crypt_flag==ENCRYPT) {
       fprintf(stderr,"Plaintext: %s\n",key);
     } else {
       DUMP_KEY(key,24);
     }
   }

   fp = fopen(MASTERKEY_FILE, "r");
   if (fp == NULL) {
       printf("\nError: can't open %s\n", MASTERKEY_FILE);
       return 0;
   }

   secret[0] = '\0';
   do {
       tmp = fgets(file_line, MAX_LINE_LEN, fp);
       if (!tmp) {
	   break;
       }
       line_count++;
       if (*tmp == '#')
	   continue;
       if (sscanf(file_line, "%s\n", secret) != 1) {
	   printf("Warning: incorrect entry format at line %d in file %s\n", line_count, MASTERKEY_FILE);
	   secret[0] = '\0';
	   continue;
       }
   } while (!secret[0]);
  
   if (!secret) {
       printf("Masterkey is not found in file %s\n", MASTERKEY_FILE);
       return 0;
   }
   bcopy(secret, md5input, OCT_KEY_LEN);
   strncat(md5input, username, strlen(username));
   md5_calc(md5output, md5input, strlen(md5input));

   if (!crypt_flag) {
       tmp = (char *) malloc(sizeof(char) * MD5_OUTPUT);	
       bcopy(key, tmp, MD5_OUTPUT);
   }
   for (i = 0; i < MD5_OUTPUT; i++) {
       key[i] ^= md5output[i];
   }

   bcopy(secret, md5input, OCT_KEY_LEN);
   md5input[strlen(secret)+1] = '\0';
   strncat(md5input, crypt_flag ? key : tmp, MD5_OUTPUT);
   md5_calc(md5output, md5input, strlen(md5input));
   for (i = 0; i < 8; i++) {
       key[MD5_OUTPUT + i] ^= md5output[i];
   }

   fclose(fp);

   if(debug_flag) {
     if(crypt_flag==DECRYPT) {
       fprintf(stderr,"Plaintext: %s\n",key);
     } else {
       DUMP_KEY(key,24);
     }
   }

   return 1;
}

